@@ -1,21 +1,29 @@
 import logging
 from pathlib import Path
+from typing import List, Optional, Union
 
 import numpy as np
 from farm.data_handler.data_silo import DataSilo
 from farm.data_handler.processor import SquadProcessor
-from farm.infer import Inferencer
+from farm.data_handler.dataloader import NamedDataLoader
+from farm.data_handler.inputs import QAInput, Question
+from farm.infer import QAInferencer
 from farm.modeling.optimization import initialize_optimizer
+from farm.modeling.predictions import QAPred, QACandidate
+from farm.modeling.adaptive_model import BaseAdaptiveModel
 from farm.train import Trainer
+from farm.eval import Evaluator
 from farm.utils import set_all_seeds, initialize_device_settings
 from scipy.special import expit
+import shutil
 
 from haystack.database.base import Document
-
+from haystack.database.elasticsearch import ElasticsearchDocumentStore
+from haystack.reader.base import BaseReader
 logger = logging.getLogger(__name__)
 
 
-class FARMReader:
+class FARMReader(BaseReader):
     """
     Transformer based model for extractive Question Answering using the FARM framework (https://github.com/deepset-ai/FARM).
     While the underlying model can vary (BERT, Roberta, DistilBERT ...) the interface remains the same.
@@ -27,16 +35,16 @@ class FARMReader:
 
     def __init__(
         self,
-        model_name_or_path,
-        context_window_size=150,
-        batch_size=50,
-        use_gpu=True,
-        no_ans_boost=None,
-        top_k_per_candidate=3,
-        top_k_per_sample=1,
-        num_processes=None,
-        max_seq_len=256,
-        doc_stride=128
+        model_name_or_path: Union[str, Path],
+        context_window_size: int = 150,
+        batch_size: int = 50,
+        use_gpu: bool = True,
+        no_ans_boost: Optional[int] = None,
+        top_k_per_candidate: int = 3,
+        top_k_per_sample: int = 1,
+        num_processes: Optional[int] = None,
+        max_seq_len: int = 256,
+        doc_stride: int = 128,
     ):
 
         """
@@ -81,7 +89,7 @@ class FARMReader:
         else:
             self.return_no_answers = True
         self.top_k_per_candidate = top_k_per_candidate
-        self.inferencer = Inferencer.load(model_name_or_path, batch_size=batch_size, gpu=use_gpu,
+        self.inferencer = QAInferencer.load(model_name_or_path, batch_size=batch_size, gpu=use_gpu,
                                           task_type="question_answering", max_seq_len=max_seq_len,
                                           doc_stride=doc_stride, num_processes=num_processes)
         self.inferencer.model.prediction_heads[0].context_window_size = context_window_size
@@ -94,9 +102,22 @@ class FARMReader:
         self.max_seq_len = max_seq_len
         self.use_gpu = use_gpu
 
-    def train(self, data_dir, train_filename, dev_filename=None, test_file_name=None,
-              use_gpu=None, batch_size=10, n_epochs=2, learning_rate=1e-5,
-              max_seq_len=None, warmup_proportion=0.2, dev_split=0.1, evaluate_every=300, save_dir=None):
+    def train(
+        self,
+        data_dir: str,
+        train_filename: str,
+        dev_filename: Optional[str] = None,
+        test_file_name: Optional[str] = None,
+        use_gpu: Optional[bool] = None,
+        batch_size: int = 10,
+        n_epochs: int = 2,
+        learning_rate: float = 1e-5,
+        max_seq_len: Optional[int] = None,
+        warmup_proportion: float = 0.2,
+        dev_split: Optional[float] = 0.1,
+        evaluate_every: int = 300,
+        save_dir: Optional[str] = None,
+    ):
         """
         Fine-tune a model on a QA dataset. Options:
         - Take a plain language model (e.g. `bert-base-cased`) and train it for QA (e.g. on SQuAD data)
@@ -138,7 +159,6 @@ class FARMReader:
 
         if not save_dir:
             save_dir = f"../../saved_models/{self.inferencer.model.language_model.name}"
-        save_dir = Path(save_dir)
 
         # 1. Create a DataProcessor that handles all the conversion from raw text into a pytorch Dataset
         label_list = ["start_token", "end_token"]
@@ -159,9 +179,17 @@ class FARMReader:
         # and calculates a few descriptive statistics of our datasets
         data_silo = DataSilo(processor=processor, batch_size=batch_size, distributed=False)
 
+        # Quick-fix until this is fixed upstream in FARM:
+        # We must avoid applying DataParallel twice (once when loading the inferencer,
+        # once when calling initalize_optimizer)
+        self.inferencer.model.save("tmp_model")
+        model = BaseAdaptiveModel.load(load_dir="tmp_model", device=device, strict=True)
+        shutil.rmtree('tmp_model')
+
         # 3. Create an optimizer and pass the already initialized model
         model, optimizer, lr_schedule = initialize_optimizer(
-            model=self.inferencer.model,
+            model=model,
+            # model=self.inferencer.model,
             learning_rate=learning_rate,
             schedule_opts={"name": "LinearWarmup", "warmup_proportion": warmup_proportion},
             n_batches=len(data_silo.loaders["train"]),
@@ -179,16 +207,18 @@ class FARMReader:
             evaluate_every=evaluate_every,
             device=device,
         )
+
+
         # 5. Let it grow!
         self.inferencer.model = trainer.train()
-        self.save(save_dir)
+        self.save(Path(save_dir))
 
-    def save(self, directory):
+    def save(self, directory: Path):
         logger.info(f"Saving reader model to {directory}")
         self.inferencer.model.save(directory)
         self.inferencer.processor.save(directory)
 
-    def predict(self, question: str, documents: [Document], top_k: int = None):
+    def predict(self, question: str, documents: List[Document], top_k: Optional[int] = None):
         """
         Use loaded QA model to find answers for a question in the supplied list of Document.
 
@@ -215,18 +245,16 @@ class FARMReader:
         """
 
         # convert input to FARM format
-        input_dicts = []
+        inputs = []
         for doc in documents:
-            cur = {
-                "text": doc.text,
-                "questions": [question],
-                "document_id": doc.id
-            }
-            input_dicts.append(cur)
+            cur = QAInput(doc_text=doc.text,
+                          questions=Question(text=question,
+                                             uid=doc.id))
+            inputs.append(cur)
 
         # get answers from QA model
-        predictions = self.inferencer.inference_from_dicts(
-            dicts=input_dicts, rest_api_schema=True, multiprocessing_chunksize=1
+        predictions = self.inferencer.inference_from_objects(
+            objects=inputs, return_json=False, multiprocessing_chunksize=1
         )
         # assemble answers from all the different documents & format them.
         # For the "no answer" option, we collect all no_ans_gaps and decide how likely
@@ -234,25 +262,28 @@ class FARMReader:
         answers = []
         no_ans_gaps = []
         best_score_answer = 0
-        for pred in predictions:
+        for pred, inp in zip(predictions, inputs):
             answers_per_document = []
-            no_ans_gaps.append(pred["predictions"][0]["no_ans_gap"])
-            for a in pred["predictions"][0]["answers"]:
+            no_ans_gaps.append(pred.no_answer_gap)
+            for ans in pred.prediction:
                 # skip "no answers" here
-                if a["answer"]:
-                    cur = {"answer": a["answer"],
-                           "score": a["score"],
-                           "probability": float(expit(np.asarray([a["score"]]) / 8)), #just a pseudo prob for now
-                           "context": a["context"],
-                           "offset_start": a["offset_answer_start"] - a["offset_context_start"],
-                           "offset_end": a["offset_answer_end"] - a["offset_context_start"],
-                           "offset_start_in_doc": a["offset_answer_start"],
-                           "offset_end_in_doc": a["offset_answer_end"],
-                           "document_id": a["document_id"]}
+                if self._check_no_answer(ans):
+                    pass
+                else:
+                    cur = {"answer": ans.answer,
+                           "score": ans.score,
+                           # just a pseudo prob for now
+                           "probability": float(expit(np.asarray([ans.score]) / 8)),  # type: ignore
+                           "context": ans.context_window,
+                           "offset_start": ans.offset_answer_start - ans.offset_context_window_start,
+                           "offset_end": ans.offset_answer_end - ans.offset_context_window_start,
+                           "offset_start_in_doc": ans.offset_answer_start,
+                           "offset_end_in_doc": ans.offset_answer_end,
+                           "document_id": pred.id}
                     answers_per_document.append(cur)
 
-                    if a["score"] > best_score_answer:
-                        best_score_answer = a["score"]
+                    if ans.score > best_score_answer:
+                        best_score_answer = ans.score
             # only take n best candidates. Answers coming back from FARM are sorted with decreasing relevance.
             answers += answers_per_document[:self.top_k_per_candidate]
 
@@ -272,8 +303,137 @@ class FARMReader:
 
         return result
 
+    def eval_on_file(self, data_dir: str, test_filename: str, device: str):
+        """
+        Performs evaluation on a SQuAD-formatted file.
+
+        Returns a dict containing the following metrics:
+            - "EM": exact match score
+            - "f1": F1-Score
+            - "top_n_accuracy": Proportion of predicted answers that match with correct answer
+
+        :param data_dir: The directory in which the test set can be found
+        :type data_dir: Path or str
+        :param test_filename: The name of the file containing the test data in SQuAD format.
+        :type test_filename: str
+        :param device: The device on which the tensors should be processed. Choose from "cpu" and "cuda".
+        :type device: str
+        """
+        eval_processor = SquadProcessor(
+            tokenizer=self.inferencer.processor.tokenizer,
+            max_seq_len=self.inferencer.processor.max_seq_len,
+            label_list=self.inferencer.processor.tasks["question_answering"]["label_list"],
+            metric=self.inferencer.processor.tasks["question_answering"]["metric"],
+            train_filename=None,
+            dev_filename=None,
+            dev_split=0,
+            test_filename=test_filename,
+            data_dir=Path(data_dir),
+        )
+
+        data_silo = DataSilo(processor=eval_processor, batch_size=self.inferencer.batch_size, distributed=False)
+        data_loader = data_silo.get_data_loader("test")
+
+        evaluator = Evaluator(data_loader=data_loader, tasks=eval_processor.tasks, device=device)
+
+        eval_results = evaluator.eval(self.inferencer.model)
+        results = {
+            "EM": eval_results[0]["EM"],
+            "f1": eval_results[0]["f1"],
+            "top_n_accuracy": eval_results[0]["top_n_accuracy"]
+        }
+        return results
+
+    def eval(
+        self,
+        document_store: ElasticsearchDocumentStore,
+        device: str,
+        label_index: str = "feedback",
+        doc_index: str = "eval_document",
+        label_origin: str = "gold_label",
+    ):
+        """
+        Performs evaluation on evaluation documents in Elasticsearch DocumentStore.
+
+        Returns a dict containing the following metrics:
+            - "EM": Proportion of exact matches of predicted answers with their corresponding correct answers
+            - "f1": Average overlap between predicted answers and their corresponding correct answers
+            - "top_n_accuracy": Proportion of predicted answers that match with correct answer
+
+        :param document_store: The ElasticsearchDocumentStore containing the evaluation documents
+        :type document_store: ElasticsearchDocumentStore
+        :param device: The device on which the tensors should be processed. Choose from "cpu" and "cuda".
+        :type device: str
+        :param label_index: Elasticsearch index where labeled questions are stored
+        :type label_index: str
+        :param doc_index: Elasticsearch index where documents that are used for evaluation are stored
+        :type doc_index: str
+        """
+
+        # extract all questions for evaluation
+        filter = {"origin": label_origin}
+        questions = document_store.get_all_documents_in_index(index=label_index, filters=filter)
+
+        # mapping from doc_id to questions
+        doc_questions_dict = {}
+        id = 0
+        for question in questions:
+            doc_id = question["_source"]["doc_id"]
+            if doc_id not in doc_questions_dict:
+                doc_questions_dict[doc_id] = [{
+                    "id": id,
+                    "question" : question["_source"]["question"],
+                    "answers" : question["_source"]["answers"],
+                    "is_impossible" : False if question["_source"]["answers"] else True
+                }]
+            else:
+                doc_questions_dict[doc_id].append({
+                    "id": id,
+                    "question" : question["_source"]["question"],
+                    "answers" : question["_source"]["answers"],
+                    "is_impossible" : False if question["_source"]["answers"] else True
+                })
+            id += 1
+
+        # extract eval documents and convert data back to SQuAD-like format
+        documents = document_store.get_all_documents_in_index(index=doc_index)
+        dicts = []
+        for document in documents:
+            doc_id = document["_source"]["doc_id"]
+            text = document["_source"]["text"]
+            questions = doc_questions_dict[doc_id]
+            dicts.append({"qas" : questions, "context" : text})
+
+        # Create DataLoader that can be passed to the Evaluator
+        indices = range(len(dicts))
+        dataset, tensor_names = self.inferencer.processor.dataset_from_dicts(dicts, indices=indices)
+        data_loader = NamedDataLoader(dataset=dataset, batch_size=self.inferencer.batch_size, tensor_names=tensor_names)
+
+        evaluator = Evaluator(data_loader=data_loader, tasks=self.inferencer.processor.tasks, device=device)
+
+        eval_results = evaluator.eval(self.inferencer.model)
+        results = {
+            "EM": eval_results[0]["EM"],
+            "f1": eval_results[0]["f1"],
+            "top_n_accuracy": eval_results[0]["top_n_accuracy"]
+        }
+        return results
+
+
     @staticmethod
-    def _calc_no_answer(no_ans_gaps,best_score_answer):
+    def _check_no_answer(c: QACandidate):
+        # check for correct value in "answer"
+        if c.offset_answer_start == 0 and c.offset_answer_end == 0:
+            if c.answer != "no_answer":
+                logger.error("Invalid 'no_answer': Got a prediction for position 0, but answer string is not 'no_answer'")
+        if c.answer == "no_answer":
+            return True
+        else:
+            return False
+
+
+    @staticmethod
+    def _calc_no_answer(no_ans_gaps: List[float], best_score_answer: float):
         # "no answer" scores and positive answers scores are difficult to compare, because
         # + a positive answer score is related to one specific document
         # - a "no answer" score is related to all input documents
@@ -283,7 +443,8 @@ class FARMReader:
         # No_ans_gap coming from FARM mean how much no_ans_boost should change to switch predictions
         no_ans_gaps = np.array(no_ans_gaps)
         max_no_ans_gap = np.max(no_ans_gaps)
-        if (np.sum(no_ans_gaps < 0) == len(no_ans_gaps)):  # all passages "no answer" as top score
+        # all passages "no answer" as top score
+        if (np.sum(no_ans_gaps < 0) == len(no_ans_gaps)):  # type: ignore
             no_ans_score = best_score_answer - max_no_ans_gap  # max_no_ans_gap is negative, so it increases best pos score
         else:  # case: at least one passage predicts an answer (positive no_ans_gap)
             no_ans_score = best_score_answer - max_no_ans_gap
@@ -297,7 +458,7 @@ class FARMReader:
                "document_id": None}
         return no_ans_prediction, max_no_ans_gap
 
-    def predict_on_texts(self, question: str, texts: [str], top_k=None):
+    def predict_on_texts(self, question: str, texts: List[str], top_k: Optional[int] = None):
         documents = []
         for i, text in enumerate(texts):
             documents.append(
@@ -308,3 +469,23 @@ class FARMReader:
             )
         predictions = self.predict(question, documents, top_k)
         return predictions
+
+    @classmethod
+    def convert_to_onnx(cls, model_name_or_path, opset_version: int = 11, optimize_for: Optional[str] = None):
+        """
+        Convert a PyTorch BERT model to ONNX format and write to ./onnx-export dir. The converted ONNX model
+        can be loaded with in the `FARMReader` using the export path as `model_name_or_path` param.
+
+        Usage:
+            >>> from haystack.reader.farm import FARMReader
+            >>> FARMReader.convert_to_onnx(model_name_or_path="deepset/bert-base-cased-squad2", optimize_for="gpu_tensor_core")
+            >>> FARMReader(model_name_or_path=Path("onnx-export"))
+
+
+        :param opset_version: ONNX opset version
+        :param optimize_for: optimize the exported model for a target device. Available options
+                             are "gpu_tensor_core" (GPUs with tensor core like V100 or T4),
+                             "gpu_without_tensor_core" (most other GPUs), and "cpu".
+        """
+        inferencer = QAInferencer.load(model_name_or_path, task_type="question_answering")
+        inferencer.model.convert_to_onnx(output_path=Path("onnx-export"), opset_version=opset_version, optimize_for=optimize_for)
