@@ -2,7 +2,7 @@ from haystack.database.base import Document
 
 
 def test_tfidf_retriever():
-    from haystack.retriever.tfidf import TfidfRetriever
+    from haystack.retriever.sparse import TfidfRetriever
 
     test_docs = [
         {"name": "testing the finder 1", "text": "godzilla says hello"},
@@ -16,4 +16,13 @@ def test_tfidf_retriever():
 
     retriever = TfidfRetriever(document_store)
     retriever.fit()
-    assert retriever.retrieve("godzilla", top_k=1) == [Document(id='0', text='godzilla says hello', external_source_id=None, question=None, query_score=None, meta=None)]
\ No newline at end of file
+    assert retriever.retrieve("godzilla", top_k=1) == [
+        Document(
+            id='26f84672c6d7aaeb8e2cd53e9c62d62d',
+            text='godzilla says hello',
+            external_source_id=None,
+            question=None,
+            query_score=None,
+            meta={"name": "testing the finder 1"},
+        )
+    ]
